from .bolt12 import (Recurrence, Bolt12, Offer, InvoiceRequest,
                     Invoice, Decoder, helper_towire_tlv, helper_fromwire_tlv,
                     simple_bech32_decode, simple_bech32_encode)
from .fundamentals import (towire_u64, towire_u32, towire_u16,
                           towire_byte, towire_tu64, towire_tu32,
                           fromwire_u64, fromwire_u32, fromwire_u16,
                           fromwire_byte, fromwire_tu64, fromwire_tu32,
                           towire_chain_hash, fromwire_chain_hash,
                           towire_sha256, fromwire_sha256,
                           towire_point32, fromwire_point32,
                           towire_point, fromwire_point,
                           towire_bip340sig, fromwire_bip340sig,
                           towire_array_utf8, fromwire_array_utf8,
                           towire_short_channel_id,
                           fromwire_short_channel_id, towire_bigsize,
                           fromwire_bigsize,
                           towire_tu16, fromwire_tu16,
                           towire_channel_id, fromwire_channel_id,
                           towire_signature, fromwire_signature)

__version__ = '0.1.2'

__all__ = ["Recurrence",
           "Bolt12",
           "Offer",
           "InvoiceRequest",
           "Invoice",
           "Decoder",
           # These are low-level, but useful if you want your own decoder
           "helper_towire_tlv",
           "helper_fromwire_tlv",
           "simple_bech32_decode",
           "simple_bech32_encode",

           "towire_u64", "towire_u32", "towire_u16",
           "towire_byte", "towire_tu64", "towire_tu32",
           "fromwire_u64", "fromwire_u32", "fromwire_u16",
           "fromwire_byte", "fromwire_tu64", "fromwire_tu32",
           "towire_chain_hash", "fromwire_chain_hash",
           "towire_sha256", "fromwire_sha256",
           "towire_point32", "fromwire_point32",
           "towire_point", "fromwire_point",
           "towire_bip340sig", "fromwire_bip340sig",
           "towire_array_utf8", "fromwire_array_utf8",
           "towire_short_channel_id",
           "fromwire_short_channel_id", "towire_bigsize",
           "fromwire_bigsize",
           "towire_tu16", "fromwire_tu16",
           "towire_channel_id", "fromwire_channel_id",
           "towire_signature", "fromwire_signature"
           ]



# Generated code from tools/generate-code.py
from .fundamentals import (towire_u64, towire_u32, towire_u16,
                           towire_byte, towire_tu64, towire_tu32,
                           fromwire_u64, fromwire_u32, fromwire_u16,
                           fromwire_byte, fromwire_tu64, fromwire_tu32,
                           towire_chain_hash, fromwire_chain_hash,
                           towire_sha256, fromwire_sha256,
                           towire_point32, fromwire_point32,
                           towire_point, fromwire_point,
                           towire_bip340sig, fromwire_bip340sig,
                           towire_array_utf8, fromwire_array_utf8,
                           towire_short_channel_id,
                           fromwire_short_channel_id, towire_bigsize,
                           fromwire_bigsize,
                           # flake-8 complains about currently unused:
                           # towire_tu16, fromwire_tu16
                           # towire_channel_id, fromwire_channel_id,
                           # towire_signature, fromwire_signature
                           )


def towire_tlv_payload_amt_to_forward(value):
    _n = 0
    buf = bytes()
    value = {"amt_to_forward": value}
    buf += towire_tu64(value["amt_to_forward"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_tlv_payload_amt_to_forward(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["amt_to_forward"] = val

    return value["amt_to_forward"], buffer


def towire_tlv_payload_outgoing_cltv_value(value):
    _n = 0
    buf = bytes()
    value = {"outgoing_cltv_value": value}
    buf += towire_tu32(value["outgoing_cltv_value"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_tlv_payload_outgoing_cltv_value(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["outgoing_cltv_value"] = val

    return value["outgoing_cltv_value"], buffer


def towire_tlv_payload_short_channel_id(value):
    _n = 0
    buf = bytes()
    value = {"short_channel_id": value}
    buf += towire_short_channel_id(value["short_channel_id"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_tlv_payload_short_channel_id(buffer):
    value = {}
    val, buffer = fromwire_short_channel_id(buffer)
    value["short_channel_id"] = val

    return value["short_channel_id"], buffer


def towire_tlv_payload_payment_data(value):
    _n = 0
    buf = bytes()
    assert len(value["payment_secret"]) == 32
    for v in value["payment_secret"]:
        buf += towire_byte(v)
    _n += 1
    buf += towire_tu64(value["total_msat"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_tlv_payload_payment_data(buffer):
    value = {}
    v = []
    i = 0
    while i < 32:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["payment_secret"] = v
    val, buffer = fromwire_tu64(buffer)
    value["total_msat"] = val

    return value, buffer


tlv_tlv_payload = {
    2: ("amt_to_forward", towire_tlv_payload_amt_to_forward, fromwire_tlv_payload_amt_to_forward),
    4: ("outgoing_cltv_value", towire_tlv_payload_outgoing_cltv_value, fromwire_tlv_payload_outgoing_cltv_value),
    6: ("short_channel_id", towire_tlv_payload_short_channel_id, fromwire_tlv_payload_short_channel_id),
    8: ("payment_data", towire_tlv_payload_payment_data, fromwire_tlv_payload_payment_data),
}


def towire_offer_chains(value):
    _n = 0
    buf = bytes()
    value = {"chains": value}
    for v in value["chains"]:
        buf += towire_chain_hash(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_chains(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_chain_hash(buffer)
        v.append(val)
        i += 1
    value["chains"] = v

    return value["chains"], buffer


def towire_offer_currency(value):
    _n = 0
    buf = bytes()
    value = {"iso4217": value}
    buf += towire_array_utf8(value["iso4217"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_currency(buffer):
    value = {}
    value["iso4217"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["iso4217"], buffer


def towire_offer_amount(value):
    _n = 0
    buf = bytes()
    value = {"amount": value}
    buf += towire_tu64(value["amount"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_amount(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["amount"] = val

    return value["amount"], buffer


def towire_offer_description(value):
    _n = 0
    buf = bytes()
    value = {"description": value}
    buf += towire_array_utf8(value["description"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_description(buffer):
    value = {}
    value["description"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["description"], buffer


def towire_offer_features(value):
    _n = 0
    buf = bytes()
    value = {"features": value}
    for v in value["features"]:
        buf += towire_byte(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_features(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["features"] = v

    return value["features"], buffer


def towire_offer_absolute_expiry(value):
    _n = 0
    buf = bytes()
    value = {"seconds_from_epoch": value}
    buf += towire_tu64(value["seconds_from_epoch"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_absolute_expiry(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["seconds_from_epoch"] = val

    return value["seconds_from_epoch"], buffer


def towire_offer_paths(value):
    _n = 0
    buf = bytes()
    value = {"paths": value}
    for v in value["paths"]:
        buf += towire_blinded_path(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_paths(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_blinded_path(buffer)
        v.append(val)
        i += 1
    value["paths"] = v

    return value["paths"], buffer


def towire_offer_vendor(value):
    _n = 0
    buf = bytes()
    value = {"vendor": value}
    buf += towire_array_utf8(value["vendor"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_vendor(buffer):
    value = {}
    value["vendor"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["vendor"], buffer


def towire_offer_quantity_min(value):
    _n = 0
    buf = bytes()
    value = {"min": value}
    buf += towire_tu64(value["min"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_quantity_min(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["min"] = val

    return value["min"], buffer


def towire_offer_quantity_max(value):
    _n = 0
    buf = bytes()
    value = {"max": value}
    buf += towire_tu64(value["max"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_quantity_max(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["max"] = val

    return value["max"], buffer


def towire_offer_recurrence(value):
    _n = 0
    buf = bytes()
    buf += towire_byte(value["time_unit"])
    _n += 1
    buf += towire_tu32(value["period"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_recurrence(buffer):
    value = {}
    val, buffer = fromwire_byte(buffer)
    value["time_unit"] = val
    val, buffer = fromwire_tu32(buffer)
    value["period"] = val

    return value, buffer


def towire_offer_recurrence_paywindow(value):
    _n = 0
    buf = bytes()
    buf += towire_u32(value["seconds_before"])
    _n += 1
    buf += towire_byte(value["proportional_amount"])
    _n += 1
    buf += towire_tu32(value["seconds_after"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_recurrence_paywindow(buffer):
    value = {}
    val, buffer = fromwire_u32(buffer)
    value["seconds_before"] = val
    val, buffer = fromwire_byte(buffer)
    value["proportional_amount"] = val
    val, buffer = fromwire_tu32(buffer)
    value["seconds_after"] = val

    return value, buffer


def towire_offer_recurrence_limit(value):
    _n = 0
    buf = bytes()
    value = {"max_period": value}
    buf += towire_tu32(value["max_period"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_recurrence_limit(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["max_period"] = val

    return value["max_period"], buffer


def towire_offer_recurrence_base(value):
    _n = 0
    buf = bytes()
    buf += towire_byte(value["start_any_period"])
    _n += 1
    buf += towire_tu64(value["basetime"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_recurrence_base(buffer):
    value = {}
    val, buffer = fromwire_byte(buffer)
    value["start_any_period"] = val
    val, buffer = fromwire_tu64(buffer)
    value["basetime"] = val

    return value, buffer


def towire_offer_node_id(value):
    _n = 0
    buf = bytes()
    value = {"node_id": value}
    buf += towire_point32(value["node_id"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_node_id(buffer):
    value = {}
    val, buffer = fromwire_point32(buffer)
    value["node_id"] = val

    return value["node_id"], buffer


def towire_offer_send_invoice(value):
    _n = 0
    buf = bytes()
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_send_invoice(buffer):
    value = {}

    return value, buffer


def towire_offer_refund_for(value):
    _n = 0
    buf = bytes()
    value = {"refunded_payment_hash": value}
    buf += towire_sha256(value["refunded_payment_hash"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_refund_for(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["refunded_payment_hash"] = val

    return value["refunded_payment_hash"], buffer


def towire_offer_signature(value):
    _n = 0
    buf = bytes()
    value = {"sig": value}
    buf += towire_bip340sig(value["sig"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_offer_signature(buffer):
    value = {}
    val, buffer = fromwire_bip340sig(buffer)
    value["sig"] = val

    return value["sig"], buffer


tlv_offer = {
    2: ("chains", towire_offer_chains, fromwire_offer_chains),
    6: ("currency", towire_offer_currency, fromwire_offer_currency),
    8: ("amount", towire_offer_amount, fromwire_offer_amount),
    10: ("description", towire_offer_description, fromwire_offer_description),
    12: ("features", towire_offer_features, fromwire_offer_features),
    14: ("absolute_expiry", towire_offer_absolute_expiry, fromwire_offer_absolute_expiry),
    16: ("paths", towire_offer_paths, fromwire_offer_paths),
    20: ("vendor", towire_offer_vendor, fromwire_offer_vendor),
    22: ("quantity_min", towire_offer_quantity_min, fromwire_offer_quantity_min),
    24: ("quantity_max", towire_offer_quantity_max, fromwire_offer_quantity_max),
    26: ("recurrence", towire_offer_recurrence, fromwire_offer_recurrence),
    64: ("recurrence_paywindow", towire_offer_recurrence_paywindow, fromwire_offer_recurrence_paywindow),
    66: ("recurrence_limit", towire_offer_recurrence_limit, fromwire_offer_recurrence_limit),
    28: ("recurrence_base", towire_offer_recurrence_base, fromwire_offer_recurrence_base),
    30: ("node_id", towire_offer_node_id, fromwire_offer_node_id),
    54: ("send_invoice", towire_offer_send_invoice, fromwire_offer_send_invoice),
    34: ("refund_for", towire_offer_refund_for, fromwire_offer_refund_for),
    240: ("signature", towire_offer_signature, fromwire_offer_signature),
}


def towire_invoice_request_chains(value):
    _n = 0
    buf = bytes()
    value = {"chains": value}
    for v in value["chains"]:
        buf += towire_chain_hash(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_chains(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_chain_hash(buffer)
        v.append(val)
        i += 1
    value["chains"] = v

    return value["chains"], buffer


def towire_invoice_request_offer_id(value):
    _n = 0
    buf = bytes()
    value = {"offer_id": value}
    buf += towire_sha256(value["offer_id"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_offer_id(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["offer_id"] = val

    return value["offer_id"], buffer


def towire_invoice_request_amount(value):
    _n = 0
    buf = bytes()
    value = {"msat": value}
    buf += towire_tu64(value["msat"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_amount(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["msat"] = val

    return value["msat"], buffer


def towire_invoice_request_features(value):
    _n = 0
    buf = bytes()
    value = {"features": value}
    for v in value["features"]:
        buf += towire_byte(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_features(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["features"] = v

    return value["features"], buffer


def towire_invoice_request_quantity(value):
    _n = 0
    buf = bytes()
    value = {"quantity": value}
    buf += towire_tu64(value["quantity"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_quantity(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["quantity"] = val

    return value["quantity"], buffer


def towire_invoice_request_recurrence_counter(value):
    _n = 0
    buf = bytes()
    value = {"counter": value}
    buf += towire_tu32(value["counter"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_recurrence_counter(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["counter"] = val

    return value["counter"], buffer


def towire_invoice_request_recurrence_start(value):
    _n = 0
    buf = bytes()
    value = {"period_offset": value}
    buf += towire_tu32(value["period_offset"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_recurrence_start(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["period_offset"] = val

    return value["period_offset"], buffer


def towire_invoice_request_payer_key(value):
    _n = 0
    buf = bytes()
    value = {"key": value}
    buf += towire_point32(value["key"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_payer_key(buffer):
    value = {}
    val, buffer = fromwire_point32(buffer)
    value["key"] = val

    return value["key"], buffer


def towire_invoice_request_payer_note(value):
    _n = 0
    buf = bytes()
    value = {"note": value}
    buf += towire_array_utf8(value["note"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_payer_note(buffer):
    value = {}
    value["note"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["note"], buffer


def towire_invoice_request_payer_info(value):
    _n = 0
    buf = bytes()
    value = {"blob": value}
    for v in value["blob"]:
        buf += towire_byte(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_payer_info(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["blob"] = v

    return value["blob"], buffer


def towire_invoice_request_replace_invoice(value):
    _n = 0
    buf = bytes()
    value = {"payment_hash": value}
    buf += towire_sha256(value["payment_hash"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_replace_invoice(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["payment_hash"] = val

    return value["payment_hash"], buffer


def towire_invoice_request_payer_signature(value):
    _n = 0
    buf = bytes()
    value = {"sig": value}
    buf += towire_bip340sig(value["sig"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_request_payer_signature(buffer):
    value = {}
    val, buffer = fromwire_bip340sig(buffer)
    value["sig"] = val

    return value["sig"], buffer


tlv_invoice_request = {
    2: ("chains", towire_invoice_request_chains, fromwire_invoice_request_chains),
    4: ("offer_id", towire_invoice_request_offer_id, fromwire_invoice_request_offer_id),
    8: ("amount", towire_invoice_request_amount, fromwire_invoice_request_amount),
    12: ("features", towire_invoice_request_features, fromwire_invoice_request_features),
    32: ("quantity", towire_invoice_request_quantity, fromwire_invoice_request_quantity),
    36: ("recurrence_counter", towire_invoice_request_recurrence_counter, fromwire_invoice_request_recurrence_counter),
    68: ("recurrence_start", towire_invoice_request_recurrence_start, fromwire_invoice_request_recurrence_start),
    38: ("payer_key", towire_invoice_request_payer_key, fromwire_invoice_request_payer_key),
    39: ("payer_note", towire_invoice_request_payer_note, fromwire_invoice_request_payer_note),
    50: ("payer_info", towire_invoice_request_payer_info, fromwire_invoice_request_payer_info),
    56: ("replace_invoice", towire_invoice_request_replace_invoice, fromwire_invoice_request_replace_invoice),
    240: ("payer_signature", towire_invoice_request_payer_signature, fromwire_invoice_request_payer_signature),
}


def towire_invoice_chains(value):
    _n = 0
    buf = bytes()
    value = {"chains": value}
    for v in value["chains"]:
        buf += towire_chain_hash(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_chains(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_chain_hash(buffer)
        v.append(val)
        i += 1
    value["chains"] = v

    return value["chains"], buffer


def towire_invoice_offer_id(value):
    _n = 0
    buf = bytes()
    value = {"offer_id": value}
    buf += towire_sha256(value["offer_id"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_offer_id(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["offer_id"] = val

    return value["offer_id"], buffer


def towire_invoice_amount(value):
    _n = 0
    buf = bytes()
    value = {"msat": value}
    buf += towire_tu64(value["msat"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_amount(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["msat"] = val

    return value["msat"], buffer


def towire_invoice_description(value):
    _n = 0
    buf = bytes()
    value = {"description": value}
    buf += towire_array_utf8(value["description"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_description(buffer):
    value = {}
    value["description"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["description"], buffer


def towire_invoice_features(value):
    _n = 0
    buf = bytes()
    value = {"features": value}
    for v in value["features"]:
        buf += towire_byte(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_features(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["features"] = v

    return value["features"], buffer


def towire_invoice_paths(value):
    _n = 0
    buf = bytes()
    value = {"paths": value}
    for v in value["paths"]:
        buf += towire_blinded_path(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_paths(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_blinded_path(buffer)
        v.append(val)
        i += 1
    value["paths"] = v

    return value["paths"], buffer


def towire_invoice_blindedpay(value):
    _n = 0
    buf = bytes()
    value = {"payinfo": value}
    for v in value["payinfo"]:
        buf += towire_blinded_payinfo(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_blindedpay(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_blinded_payinfo(buffer)
        v.append(val)
        i += 1
    value["payinfo"] = v

    return value["payinfo"], buffer


def towire_invoice_blinded_capacities(value):
    _n = 0
    buf = bytes()
    value = {"incoming_msat": value}
    for v in value["incoming_msat"]:
        buf += towire_u64(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_blinded_capacities(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_u64(buffer)
        v.append(val)
        i += 1
    value["incoming_msat"] = v

    return value["incoming_msat"], buffer


def towire_invoice_vendor(value):
    _n = 0
    buf = bytes()
    value = {"vendor": value}
    buf += towire_array_utf8(value["vendor"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_vendor(buffer):
    value = {}
    value["vendor"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["vendor"], buffer


def towire_invoice_node_id(value):
    _n = 0
    buf = bytes()
    value = {"node_id": value}
    buf += towire_point32(value["node_id"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_node_id(buffer):
    value = {}
    val, buffer = fromwire_point32(buffer)
    value["node_id"] = val

    return value["node_id"], buffer


def towire_invoice_quantity(value):
    _n = 0
    buf = bytes()
    value = {"quantity": value}
    buf += towire_tu64(value["quantity"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_quantity(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["quantity"] = val

    return value["quantity"], buffer


def towire_invoice_refund_for(value):
    _n = 0
    buf = bytes()
    value = {"refunded_payment_hash": value}
    buf += towire_sha256(value["refunded_payment_hash"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_refund_for(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["refunded_payment_hash"] = val

    return value["refunded_payment_hash"], buffer


def towire_invoice_recurrence_counter(value):
    _n = 0
    buf = bytes()
    value = {"counter": value}
    buf += towire_tu32(value["counter"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_recurrence_counter(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["counter"] = val

    return value["counter"], buffer


def towire_invoice_send_invoice(value):
    _n = 0
    buf = bytes()
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_send_invoice(buffer):
    value = {}

    return value, buffer


def towire_invoice_recurrence_start(value):
    _n = 0
    buf = bytes()
    value = {"period_offset": value}
    buf += towire_tu32(value["period_offset"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_recurrence_start(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["period_offset"] = val

    return value["period_offset"], buffer


def towire_invoice_recurrence_basetime(value):
    _n = 0
    buf = bytes()
    value = {"basetime": value}
    buf += towire_tu64(value["basetime"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_recurrence_basetime(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["basetime"] = val

    return value["basetime"], buffer


def towire_invoice_payer_key(value):
    _n = 0
    buf = bytes()
    value = {"key": value}
    buf += towire_point32(value["key"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_payer_key(buffer):
    value = {}
    val, buffer = fromwire_point32(buffer)
    value["key"] = val

    return value["key"], buffer


def towire_invoice_payer_note(value):
    _n = 0
    buf = bytes()
    value = {"note": value}
    buf += towire_array_utf8(value["note"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_payer_note(buffer):
    value = {}
    value["note"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["note"], buffer


def towire_invoice_payer_info(value):
    _n = 0
    buf = bytes()
    value = {"blob": value}
    for v in value["blob"]:
        buf += towire_byte(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_payer_info(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["blob"] = v

    return value["blob"], buffer


def towire_invoice_created_at(value):
    _n = 0
    buf = bytes()
    value = {"timestamp": value}
    buf += towire_tu64(value["timestamp"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_created_at(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["timestamp"] = val

    return value["timestamp"], buffer


def towire_invoice_payment_hash(value):
    _n = 0
    buf = bytes()
    value = {"payment_hash": value}
    buf += towire_sha256(value["payment_hash"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_payment_hash(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["payment_hash"] = val

    return value["payment_hash"], buffer


def towire_invoice_relative_expiry(value):
    _n = 0
    buf = bytes()
    value = {"seconds_from_creation": value}
    buf += towire_tu32(value["seconds_from_creation"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_relative_expiry(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["seconds_from_creation"] = val

    return value["seconds_from_creation"], buffer


def towire_invoice_cltv(value):
    _n = 0
    buf = bytes()
    value = {"min_final_cltv_expiry": value}
    buf += towire_tu32(value["min_final_cltv_expiry"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_cltv(buffer):
    value = {}
    val, buffer = fromwire_tu32(buffer)
    value["min_final_cltv_expiry"] = val

    return value["min_final_cltv_expiry"], buffer


def towire_invoice_fallbacks(value):
    _n = 0
    buf = bytes()
    buf += towire_byte(len(value["fallbacks"]))
    _n += 1
    for v in value["fallbacks"]:
        buf += towire_fallback_address(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_fallbacks(buffer):
    value = {}
    lenfield_num = fromwire_byte(buffer)
    v = []
    i = 0
    while i < lenfield_num:
        val, buffer = fromwire_fallback_address(buffer)
        v.append(val)
        i += 1
    value["fallbacks"] = v

    return value, buffer


def towire_invoice_refund_signature(value):
    _n = 0
    buf = bytes()
    value = {"payer_signature": value}
    buf += towire_bip340sig(value["payer_signature"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_refund_signature(buffer):
    value = {}
    val, buffer = fromwire_bip340sig(buffer)
    value["payer_signature"] = val

    return value["payer_signature"], buffer


def towire_invoice_replace_invoice(value):
    _n = 0
    buf = bytes()
    value = {"payment_hash": value}
    buf += towire_sha256(value["payment_hash"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_replace_invoice(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["payment_hash"] = val

    return value["payment_hash"], buffer


def towire_invoice_signature(value):
    _n = 0
    buf = bytes()
    value = {"sig": value}
    buf += towire_bip340sig(value["sig"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_signature(buffer):
    value = {}
    val, buffer = fromwire_bip340sig(buffer)
    value["sig"] = val

    return value["sig"], buffer


tlv_invoice = {
    2: ("chains", towire_invoice_chains, fromwire_invoice_chains),
    4: ("offer_id", towire_invoice_offer_id, fromwire_invoice_offer_id),
    8: ("amount", towire_invoice_amount, fromwire_invoice_amount),
    10: ("description", towire_invoice_description, fromwire_invoice_description),
    12: ("features", towire_invoice_features, fromwire_invoice_features),
    16: ("paths", towire_invoice_paths, fromwire_invoice_paths),
    18: ("blindedpay", towire_invoice_blindedpay, fromwire_invoice_blindedpay),
    19: ("blinded_capacities", towire_invoice_blinded_capacities, fromwire_invoice_blinded_capacities),
    20: ("vendor", towire_invoice_vendor, fromwire_invoice_vendor),
    30: ("node_id", towire_invoice_node_id, fromwire_invoice_node_id),
    32: ("quantity", towire_invoice_quantity, fromwire_invoice_quantity),
    34: ("refund_for", towire_invoice_refund_for, fromwire_invoice_refund_for),
    36: ("recurrence_counter", towire_invoice_recurrence_counter, fromwire_invoice_recurrence_counter),
    54: ("send_invoice", towire_invoice_send_invoice, fromwire_invoice_send_invoice),
    68: ("recurrence_start", towire_invoice_recurrence_start, fromwire_invoice_recurrence_start),
    64: ("recurrence_basetime", towire_invoice_recurrence_basetime, fromwire_invoice_recurrence_basetime),
    38: ("payer_key", towire_invoice_payer_key, fromwire_invoice_payer_key),
    39: ("payer_note", towire_invoice_payer_note, fromwire_invoice_payer_note),
    50: ("payer_info", towire_invoice_payer_info, fromwire_invoice_payer_info),
    40: ("created_at", towire_invoice_created_at, fromwire_invoice_created_at),
    42: ("payment_hash", towire_invoice_payment_hash, fromwire_invoice_payment_hash),
    44: ("relative_expiry", towire_invoice_relative_expiry, fromwire_invoice_relative_expiry),
    46: ("cltv", towire_invoice_cltv, fromwire_invoice_cltv),
    48: ("fallbacks", towire_invoice_fallbacks, fromwire_invoice_fallbacks),
    52: ("refund_signature", towire_invoice_refund_signature, fromwire_invoice_refund_signature),
    56: ("replace_invoice", towire_invoice_replace_invoice, fromwire_invoice_replace_invoice),
    240: ("signature", towire_invoice_signature, fromwire_invoice_signature),
}


def towire_invoice_error_erroneous_field(value):
    _n = 0
    buf = bytes()
    value = {"tlv_fieldnum": value}
    buf += towire_tu64(value["tlv_fieldnum"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_error_erroneous_field(buffer):
    value = {}
    val, buffer = fromwire_tu64(buffer)
    value["tlv_fieldnum"] = val

    return value["tlv_fieldnum"], buffer


def towire_invoice_error_suggested_value(value):
    _n = 0
    buf = bytes()
    value = {"value": value}
    for v in value["value"]:
        buf += towire_byte(v)
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_error_suggested_value(buffer):
    value = {}
    v = []
    i = 0
    while len(buffer) != 0:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["value"] = v

    return value["value"], buffer


def towire_invoice_error_error(value):
    _n = 0
    buf = bytes()
    value = {"msg": value}
    buf += towire_array_utf8(value["msg"])
    _n += 1
    # Ensures there are no extra keys!
    assert len(value) == _n
    return buf


def fromwire_invoice_error_error(buffer):
    value = {}
    value["msg"], buffer = fromwire_array_utf8(buffer, len(buffer))

    return value["msg"], buffer


tlv_invoice_error = {
    1: ("erroneous_field", towire_invoice_error_erroneous_field, fromwire_invoice_error_erroneous_field),
    3: ("suggested_value", towire_invoice_error_suggested_value, fromwire_invoice_error_suggested_value),
    5: ("error", towire_invoice_error_error, fromwire_invoice_error_error),
}


def towire_onionmsg_path(value):
    _n = 0
    buf = bytes()
    buf += towire_point(value["node_id"])
    _n += 1
    buf += towire_u16(len(value["enctlv"]))
    _n += 1
    for v in value["enctlv"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_onionmsg_path(buffer):
    value = {}
    val, buffer = fromwire_point(buffer)
    value["node_id"] = val
    lenfield_enclen = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_enclen:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["enctlv"] = v

    return value, buffer


def towire_blinded_path(value):
    _n = 0
    buf = bytes()
    buf += towire_point(value["blinding"])
    _n += 1
    buf += towire_u16(len(value["path"]))
    _n += 1
    for v in value["path"]:
        buf += towire_onionmsg_path(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_blinded_path(buffer):
    value = {}
    val, buffer = fromwire_point(buffer)
    value["blinding"] = val
    lenfield_num_hops = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_num_hops:
        val, buffer = fromwire_onionmsg_path(buffer)
        v.append(val)
        i += 1
    value["path"] = v

    return value, buffer


def towire_blinded_payinfo(value):
    _n = 0
    buf = bytes()
    buf += towire_u32(value["fee_base_msat"])
    _n += 1
    buf += towire_u32(value["fee_proportional_millionths"])
    _n += 1
    buf += towire_u16(value["cltv_expiry_delta"])
    _n += 1
    buf += towire_u16(len(value["features"]))
    _n += 1
    for v in value["features"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_blinded_payinfo(buffer):
    value = {}
    val, buffer = fromwire_u32(buffer)
    value["fee_base_msat"] = val
    val, buffer = fromwire_u32(buffer)
    value["fee_proportional_millionths"] = val
    val, buffer = fromwire_u16(buffer)
    value["cltv_expiry_delta"] = val
    lenfield_flen = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_flen:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["features"] = v

    return value, buffer


def towire_fallback_address(value):
    _n = 0
    buf = bytes()
    buf += towire_byte(value["version"])
    _n += 1
    buf += towire_u16(len(value["address"]))
    _n += 1
    for v in value["address"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_fallback_address(buffer):
    value = {}
    val, buffer = fromwire_byte(buffer)
    value["version"] = val
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["address"] = v

    return value, buffer


def towire_invalid_realm(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_invalid_realm(buffer):
    value = {}

    return value, buffer


def towire_temporary_node_failure(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_temporary_node_failure(buffer):
    value = {}

    return value, buffer


def towire_permanent_node_failure(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_permanent_node_failure(buffer):
    value = {}

    return value, buffer


def towire_required_node_feature_missing(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_required_node_feature_missing(buffer):
    value = {}

    return value, buffer


def towire_invalid_onion_version(value):
    _n = 0
    buf = bytes()
    buf += towire_sha256(value["sha256_of_onion"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_invalid_onion_version(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["sha256_of_onion"] = val

    return value, buffer


def towire_invalid_onion_hmac(value):
    _n = 0
    buf = bytes()
    buf += towire_sha256(value["sha256_of_onion"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_invalid_onion_hmac(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["sha256_of_onion"] = val

    return value, buffer


def towire_invalid_onion_key(value):
    _n = 0
    buf = bytes()
    buf += towire_sha256(value["sha256_of_onion"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_invalid_onion_key(buffer):
    value = {}
    val, buffer = fromwire_sha256(buffer)
    value["sha256_of_onion"] = val

    return value, buffer


def towire_temporary_channel_failure(value):
    _n = 0
    buf = bytes()
    buf += towire_u16(len(value["channel_update"]))
    _n += 1
    for v in value["channel_update"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_temporary_channel_failure(buffer):
    value = {}
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["channel_update"] = v

    return value, buffer


def towire_permanent_channel_failure(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_permanent_channel_failure(buffer):
    value = {}

    return value, buffer


def towire_required_channel_feature_missing(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_required_channel_feature_missing(buffer):
    value = {}

    return value, buffer


def towire_unknown_next_peer(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_unknown_next_peer(buffer):
    value = {}

    return value, buffer


def towire_amount_below_minimum(value):
    _n = 0
    buf = bytes()
    buf += towire_u64(value["htlc_msat"])
    _n += 1
    buf += towire_u16(len(value["channel_update"]))
    _n += 1
    for v in value["channel_update"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_amount_below_minimum(buffer):
    value = {}
    val, buffer = fromwire_u64(buffer)
    value["htlc_msat"] = val
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["channel_update"] = v

    return value, buffer


def towire_fee_insufficient(value):
    _n = 0
    buf = bytes()
    buf += towire_u64(value["htlc_msat"])
    _n += 1
    buf += towire_u16(len(value["channel_update"]))
    _n += 1
    for v in value["channel_update"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_fee_insufficient(buffer):
    value = {}
    val, buffer = fromwire_u64(buffer)
    value["htlc_msat"] = val
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["channel_update"] = v

    return value, buffer


def towire_incorrect_cltv_expiry(value):
    _n = 0
    buf = bytes()
    buf += towire_u32(value["cltv_expiry"])
    _n += 1
    buf += towire_u16(len(value["channel_update"]))
    _n += 1
    for v in value["channel_update"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_incorrect_cltv_expiry(buffer):
    value = {}
    val, buffer = fromwire_u32(buffer)
    value["cltv_expiry"] = val
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["channel_update"] = v

    return value, buffer


def towire_expiry_too_soon(value):
    _n = 0
    buf = bytes()
    buf += towire_u16(len(value["channel_update"]))
    _n += 1
    for v in value["channel_update"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_expiry_too_soon(buffer):
    value = {}
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["channel_update"] = v

    return value, buffer


def towire_incorrect_or_unknown_payment_details(value):
    _n = 0
    buf = bytes()
    buf += towire_u64(value["htlc_msat"])
    _n += 1
    buf += towire_u32(value["height"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_incorrect_or_unknown_payment_details(buffer):
    value = {}
    val, buffer = fromwire_u64(buffer)
    value["htlc_msat"] = val
    val, buffer = fromwire_u32(buffer)
    value["height"] = val

    return value, buffer


def towire_final_incorrect_cltv_expiry(value):
    _n = 0
    buf = bytes()
    buf += towire_u32(value["cltv_expiry"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_final_incorrect_cltv_expiry(buffer):
    value = {}
    val, buffer = fromwire_u32(buffer)
    value["cltv_expiry"] = val

    return value, buffer


def towire_final_incorrect_htlc_amount(value):
    _n = 0
    buf = bytes()
    buf += towire_u64(value["incoming_htlc_amt"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_final_incorrect_htlc_amount(buffer):
    value = {}
    val, buffer = fromwire_u64(buffer)
    value["incoming_htlc_amt"] = val

    return value, buffer


def towire_channel_disabled(value):
    _n = 0
    buf = bytes()
    buf += towire_u16(value["flags"])
    _n += 1
    buf += towire_u16(len(value["channel_update"]))
    _n += 1
    for v in value["channel_update"]:
        buf += towire_byte(v)
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_channel_disabled(buffer):
    value = {}
    val, buffer = fromwire_u16(buffer)
    value["flags"] = val
    lenfield_len = fromwire_u16(buffer)
    v = []
    i = 0
    while i < lenfield_len:
        val, buffer = fromwire_byte(buffer)
        v.append(val)
        i += 1
    value["channel_update"] = v

    return value, buffer


def towire_expiry_too_far(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_expiry_too_far(buffer):
    value = {}

    return value, buffer


def towire_invalid_onion_payload(value):
    _n = 0
    buf = bytes()
    buf += towire_bigsize(value["type"])
    _n += 1
    buf += towire_u16(value["offset"])
    _n += 1
    assert len(value) == _n
    return buf


def fromwire_invalid_onion_payload(buffer):
    value = {}
    val, buffer = fromwire_bigsize(buffer)
    value["type"] = val
    val, buffer = fromwire_u16(buffer)
    value["offset"] = val

    return value, buffer


def towire_mpp_timeout(value):
    _n = 0
    buf = bytes()
    assert len(value) == _n
    return buf


def fromwire_mpp_timeout(buffer):
    value = {}

    return value, buffer



# Generated code from tools/generate-code.py
from .fundamentals import (towire_u64, towire_u32, towire_u16,
                           towire_byte, towire_tu64, towire_tu32,
                           fromwire_u64, fromwire_u32, fromwire_u16,
                           fromwire_byte, fromwire_tu64, fromwire_tu32,
                           towire_chain_hash, fromwire_chain_hash,
                           towire_sha256, fromwire_sha256,
                           towire_point32, fromwire_point32,
                           towire_point, fromwire_point,
                           towire_bip340sig, fromwire_bip340sig,
                           towire_array_utf8, fromwire_array_utf8,
                           towire_short_channel_id,
                           fromwire_short_channel_id, towire_bigsize,
                           fromwire_bigsize,
                           # flake-8 complains about currently unused:
                           # towire_tu16, fromwire_tu16
                           # towire_channel_id, fromwire_channel_id,
                           # towire_signature, fromwire_signature
                           )



#! /usr/bin/env python3
from typing import Tuple


# Fundamental type routines
def helper_towire_int(value: int, size) -> bytes:
    return value.to_bytes(size, 'big')


def helper_towire_truncated_int(value: int, size: int) -> bytes:
    bval = value.to_bytes(size, 'big')
    while bval.startswith(b'\0'):
        bval = bval[1:]
    return bval


def helper_towire_bytes(value: bytes, size: int) -> bytes:
    if len(value) != size:
        raise ValueError("value 0x{} should be length {}"
                         .format(value.hex(), size))
    return value


def helper_fromwire_int(buffer: bytes, size: int) -> Tuple[int, bytes]:
    if len(buffer) < size:
        raise ValueError("truncated")
    return int.from_bytes(buffer[:size], 'big'), buffer[size:]


def helper_fromwire_truncated_int(buffer: bytes, size: int) -> Tuple[int, bytes]:
    if len(buffer) > size:
        raise ValueError("extra data at end")
    if buffer.startswith(b'\0'):
        raise ValueError("non-minimal tu64")
    return int.from_bytes(buffer, 'big'), bytes()


def helper_fromwire_bytes(buffer: bytes, size: int) -> Tuple[bytes, bytes]:
    if len(buffer) < size:
        raise ValueError("truncated")
    return buffer[:size], buffer[size:]


# Now the actual fundamental types:
def towire_u64(value: int) -> bytes:
    return helper_towire_int(value, 8)


def towire_u32(value: int) -> bytes:
    return helper_towire_int(value, 4)


def towire_u16(value: int) -> bytes:
    return helper_towire_int(value, 2)


def towire_byte(value: int) -> bytes:
    return helper_towire_int(value, 1)


def towire_tu64(value: int) -> bytes:
    return helper_towire_truncated_int(value, 8)


def towire_tu32(value: int) -> bytes:
    return helper_towire_truncated_int(value, 4)


def towire_tu16(value: int) -> bytes:
    return helper_towire_truncated_int(value, 2)


def fromwire_u64(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_int(buffer, 8)


def fromwire_u32(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_int(buffer, 4)


def fromwire_u16(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_int(buffer, 2)


def fromwire_byte(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_int(buffer, 1)


def fromwire_tu64(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_truncated_int(buffer, 8)


def fromwire_tu32(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_truncated_int(buffer, 4)


def fromwire_tu16(buffer: bytes) -> Tuple[int, bytes]:
    return helper_fromwire_truncated_int(buffer, 2)


def towire_chain_hash(value: bytes) -> bytes:
    return helper_towire_bytes(value, 32)


def fromwire_chain_hash(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 32)


def towire_channel_id(value: bytes) -> bytes:
    return helper_towire_bytes(value, 32)


def fromwire_channel_id(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 32)


def towire_sha256(value: bytes) -> bytes:
    return helper_towire_bytes(value, 32)


def fromwire_sha256(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 32)


def towire_point32(value: bytes) -> bytes:
    return helper_towire_bytes(value, 32)


def fromwire_point32(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 32)


def towire_point(value: bytes) -> bytes:
    return helper_towire_bytes(value, 33)


def fromwire_point(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 33)


def towire_signature(value: bytes) -> bytes:
    return helper_towire_bytes(value, 64)


def fromwire_signature(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 64)


def towire_bip340sig(value: bytes) -> bytes:
    return helper_towire_bytes(value, 64)


def fromwire_bip340sig(buffer: bytes) -> Tuple[bytes, bytes]:
    return helper_fromwire_bytes(buffer, 64)


def towire_array_utf8(value: str) -> bytes:
    return bytes(value, encoding='utf8')


def fromwire_array_utf8(buffer: bytes, size: int) -> Tuple[str, bytes]:
    bval, buffer = helper_fromwire_bytes(buffer, size)
    try:
        return bval.decode(), buffer
    except UnicodeDecodeError as e:
        raise ValueError("Bad UTF-8 at {}-{} of 0x{}"
                         .format(e.start, e.end, bval.hex()))


def towire_short_channel_id(value: str) -> bytes:
    parts = value.split('x')
    return (int(parts[0]).to_bytes(3, 'big')
            + int(parts[1]).to_bytes(3, 'big')
            + int(parts[2]).to_bytes(2, 'big'))


def fromwire_short_channel_id(buffer: bytes) -> Tuple[str, bytes]:
    intval, buffer = fromwire_u64(buffer)
    return '{}x{}x{}'.format(intval >> 48,
                             (intval >> 24) & 0xFFFFFF,
                             intval & 0xFFFF), buffer


def towire_bigsize(value: int) -> bytes:
    if value < 0xFD:
        return towire_byte(value)
    elif value <= 0xFFFF:
        return towire_byte(0xFD) + towire_u16(value)
    elif value <= 0xFFFFFFFF:
        return towire_byte(0xFE) + towire_u32(value)
    else:
        return towire_byte(0xFF) + towire_u64(value)


def fromwire_bigsize(buffer: bytes) -> Tuple[int, bytes]:
    val, buffer = fromwire_byte(buffer)
    if val == 0xFD:
        minval = 0xFD
        val, buffer = fromwire_u16(buffer)
    elif val == 0xFE:
        minval = 0x10000
        val, buffer = fromwire_u32(buffer)
    elif val == 0xFF:
        minval = 0x100000000
        val, buffer = fromwire_u64(buffer)
    else:
        minval = 0

    if val < minval:
        raise ValueError("non minimal-encoded bigsize")
    return val, buffer



#! /usr/bin/env python3
from typing import Tuple, Optional, Dict, Any, Sequence, List
import bech32
from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta
from collections import namedtuple
import hashlib
import re
import time
from .fundamentals import fromwire_bigsize, towire_bigsize
from .key import verify_schnorr, sign_schnorr
from .generated import tlv_offer, tlv_invoice_request, tlv_invoice


# BOLT #12:
# All signatures are created as per
# [BIP-340](https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki),
# and tagged as recommended there.  Thus we define H(`tag`,`msg`) as
# SHA256(SHA256(`tag`) || SHA256(`tag`) || `msg`), and SIG(`tag`,`msg`,`key`)
# as the signature of H(`tag`,`msg`) using `key`.

# FIXME: Returns hashlib.HASH?
def bolt12_h(tag: bytes, data: bytes = bytes()) -> Any:
    tag_sha = hashlib.sha256(tag).digest()
    ret = hashlib.sha256(tag_sha + tag_sha)
    ret.update(data)
    return ret


# Table-driven tlv decoder
def helper_fromwire_tlv(tlv_table: Dict[int, Tuple[str, Any, Any]],
                        buffer: bytes) -> Tuple[Dict[str, Any], Dict[int, bytes]]:
    tlvs = {}
    unknowns = {}
    prev_tlvtype = -1

    while len(buffer) != 0:
        tlv_type, buffer = fromwire_bigsize(buffer)
        if tlv_type <= prev_tlvtype:
            raise ValueError('Invalid: field number {} after {}'.format(tlv_type, prev_tlvtype))
        prev_tlvtype = tlv_type
        tlv_len, buffer = fromwire_bigsize(buffer)
        tlv_bytes = buffer[:tlv_len]
        buffer = buffer[tlv_len:]
        if len(tlv_bytes) < tlv_len:
            raise ValueError('Incomplete: field number {} truncated'.format(tlv_type))

        if tlv_type in tlv_table:
            name, _, fromwire = tlv_table[tlv_type]
            tlvs[name], tlv_bytes = fromwire(tlv_bytes)
            if len(tlv_bytes) != 0:
                raise ValueError('Invalid: field number {} extra bytes'.format(tlv_type))
        elif tlv_type % 2 == 1:
            unknowns[tlv_type] = tlv_bytes
        else:
            raise ValueError('Invalid: Unknown even field number {}'.format(tlv_type))

    return tlvs, unknowns


def tlv_ordered(tlv_table: Dict[int, Tuple[str, Any, Any]],
                tlvs: Dict[str, Any], unknowns: Dict[int, bytes],
                exclude_sigs: bool = False) -> List[Tuple[int, bytes]]:
    """Marshall all entries into bytes, arrange in increasing order"""
    def find_tlv_byname(tlv_table, tlvname):
        for tlv in tlv_table:
            if tlv_table[tlv][0] == tlvname:
                return tlv
        raise ValueError("Unknown value {} set".format(tlvname))

    # add knowns into tlv_bytestr
    tlv_bytestr = unknowns.copy()
    for tlvname, value in tlvs.items():
        tlvnum = find_tlv_byname(tlv_table, tlvname)
        if tlvnum in tlv_bytestr:
            raise ValueError("Value {} ({}) set in unknowns and tlvs".format(tlvnum, tlvname))
        tlv_bytestr[tlvnum] = tlv_table[tlvnum][1](value)

    ret = []
    for k in sorted(tlv_bytestr):
        # BOLT #12:
        # TLV types 240 through 1000 are considered signature elements.
        if exclude_sigs and k >= 240 and k <= 1000:
            continue
        ret += [(k, tlv_bytestr[k])]
    return ret


def tlv_enc(num: int, val: bytes) -> bytes:
    return towire_bigsize(num) + towire_bigsize(len(val)) + val


# Table-driven tlv encode
def helper_towire_tlv(tlv_table: Dict[int, Tuple[str, Any, Any]],
                      tlvs: Dict[str, Any],
                      unknowns: Dict[int, bytes]) -> bytes:
    buffer = bytes()
    for k, bytestr in tlv_ordered(tlv_table, tlvs, unknowns):
        buffer += tlv_enc(k, bytestr)

    return buffer


def simple_bech32_decode(bech32str: str) -> Tuple[str, bytes]:
    """Bech32 decode without a checksum"""
    # We only lower the case if ALL CAPS
    if bech32str.isupper():
        bech32str = bech32str.lower()

    sep = bech32str.find('1')
    if sep == -1:
        raise ValueError("No separator found")

    hrp = bech32str[:sep]
    if not hrp.islower():
        raise ValueError("Mixed case found")

    ret5 = bytes()
    for c in bech32str[sep + 1:]:
        pos = bech32.CHARSET.find(c)
        if pos == -1:
            raise ValueError("Invalid character '{}'".format(c))
        ret5 += bytes([pos])

    intarr = bech32.convertbits(ret5, 5, 8, False)
    # Can't fail, since we ensured all values are 0-31 above!
    assert intarr is not None
    return hrp, bytes(intarr)


def simple_bech32_encode(hrp: str, data: bytes) -> str:
    """Bech32 encode without a checksum"""
    encstr = hrp + '1'
    bits = bech32.convertbits(data, 8, 5, True)
    assert bits is not None
    for c in bits:
        encstr += bech32.CHARSET[c]
    return encstr


class Recurrence(object):
    """A class representing the period details for a recurring offer"""
    Period = namedtuple('Period', ['start', 'end', 'paystart', 'payend'])

    def __init__(self, recurrence, recurrence_paywindow=None, recurrence_limit=None, recurrence_base=None):
        self.time_unit = recurrence['time_unit']
        self.period = recurrence['period']

        self.paywindow = recurrence_paywindow
        self.limit = recurrence_limit
        self.base = recurrence_base

    def has_fixed_base(self) -> bool:
        """Does this recurrence have a fixed base, or is it from first
        invoice?"""
        return self.base is not None

    def can_start_offset(self) -> bool:
        """Does this recurrence have a fixed base, and allows start at any
period??"""
        return self.has_fixed_base() and self.base['start_any_period']

    @staticmethod
    def _adjust_date(basedate, units, n, sameday):
        while True:
            ret = basedate + relativedelta(**{units: n})
            if not sameday or ret.day == basedate.day:
                return ret

            # BOLT #12:
            # - if the day is not within the month, use the last day within
            #   the month.

            # Try one day earlier as a base
            basedate -= relativedelta(days=1)

    def _get_period(self, n: int, basetime: int) -> 'Recurrence.Period':
        """Return info on period n, ignoring limits"""
        basedate = datetime.fromtimestamp(basetime, tz=timezone.utc)

        # BOLT #12:
        # 1. A `time_unit` defining 0 (seconds), 1 (days), 2 (months),
        #    3 (years).
        # 2. A `period`, defining how often (in `time_unit`) it has to be
        #    paid.
        if self.time_unit == 0:
            units = 'seconds'
            sameday = False
        elif self.time_unit == 1:
            units = 'days'
            sameday = False
        elif self.time_unit == 2:
            units = 'months'
            sameday = True
        elif self.time_unit == 3:
            units = 'years'
            sameday = True

        startdate = self._adjust_date(basedate, units, self.period * n, sameday)
        enddate = self._adjust_date(startdate, units, self.period, sameday)

        # Convert back to UNIX seconds
        start = int(startdate.timestamp())
        end = int(enddate.timestamp())

        if self.paywindow is None:
            # Default is that you can pay during previous period or this one.
            paystartdate = self._adjust_date(startdate, units, -self.period,
                                             sameday)
            paystart = int(paystartdate.timestamp())
            payend = end
        else:
            paystart = start - self.paywindow['seconds_before']
            payend = start + self.paywindow['seconds_after']

        # You can pay this between paystart and payend.
        return Recurrence.Period(start, end, paystart, payend)

    def get_period(self,
                   n: int,
                   basetime: int = int(time.time())) -> Optional['Recurrence.Period']:
        """Return the (start,end,paystart,payend) of period n, using the
        basetime given as first period start (unless has_fixed_base).
        Returns None if that's past the limit.

        The start,end reflect when the period is, and paystart,payend reflect
        when you can pay for it.

        """
        if self.limit is not None and n > self.limit:
            return None
        if self.base is not None:
            basetime = self.base['basetime']

        return self._get_period(n, basetime)

    def get_pay_factor(self, period: 'Recurrence.Period', time=int(time.time())):
        """Returns factor by much your amount should be altered at time
        seconds; normally returns 1, but if an offer sets
        paywindow.proportional_amount, returns less than 1 if we're
        already into the period.

        """
        # If there's no proportional payment, we pay full amount
        if self.paywindow is None or not self.paywindow['proportional_amount']:
            return 1
        # Before period starts, we're good.
        if time < period.start:
            return 1
        # This can happen in theory, if paywindow goes beyond end of period.
        if time > period.end:
            return 0
        return (period.end - time) / (period.end - period.start)

    def period_start_offset(self, when: int = int(time.time())):
        """For a can_start_offset() Recurrence, what period does @when fall in?
        Can return negtive if before start!
        """
        assert self.can_start_offset()

        # Time is slippery!  Estimate using seconds, refine using exact
        # calculations (which cover leap years, etc etc etc)
        if self.time_unit == 0:
            approx_mul = 1
        elif self.time_unit == 1:
            approx_mul = 24 * 60 * 60
        elif self.time_unit == 2:
            approx_mul = 30 * 24 * 60 * 60
        elif self.time_unit == 3:
            approx_mul = 365 * 30 * 24 * 60 * 60

        period_num = ((when - self.base['basetime'])
                      // (self.period * approx_mul))

        while True:
            period = self._get_period(period_num, self.base['basetime'])
            if when < period.start:
                period_num -= 1
            elif when >= period.end:
                period_num += 1
            else:
                return period_num


class Bolt12(object):
    def __init__(self, hrp: str, tlv_table: Dict[int, Tuple[str, Any, Any]], bytestr: bytes):
        self.hrp = hrp
        self.tlv_table = tlv_table
        self.values, self.unknowns = helper_fromwire_tlv(tlv_table, bytestr)

    def encode(self) -> str:
        """Turn the BOLT12 object back into a bech32 string"""
        return simple_bech32_encode(self.hrp,
                                    helper_towire_tlv(self.tlv_table,
                                                      self.values,
                                                      self.unknowns))

    @staticmethod
    def create(hrp: str, bytestr: bytes):
        if hrp == 'lno':
            return Offer(hrp, bytestr)
        elif hrp == 'lnr':
            return InvoiceRequest(hrp, bytestr)
        elif hrp == 'lni':
            return Invoice(hrp, bytestr)

        raise ValueError('Unknown human readable prefix {}'.format(hrp))

    # FIXME: Returns 'hashlib.HASH'?
    @staticmethod
    def lnall_ctx(tlv_ordered_nosigs: List[Tuple[int, bytes]]) -> Any:
        # BOLT #12:
        # 2. The H(`LnAll`|all-tlvs,tlv) where "all-tlvs" consists of all non-signature TLV entries
        #    appended in ascending order.
        all_tlvs = bytes()
        for num, bytestr in tlv_ordered_nosigs:
            all_tlvs += tlv_enc(num, bytestr)

        return bolt12_h(b'LnAll' + all_tlvs)

    @staticmethod
    def merkle_pair(a: bytes, b: bytes) -> bytes:
        # BOLT #12:
        # The Merkle tree inner nodes are H(`LnBranch`, lesser-SHA256|greater-SHA256)
        if a < b:
            return bolt12_h(b'LnBranch', a + b).digest()
        else:
            return bolt12_h(b'LnBranch', b + a).digest()

    @staticmethod
    def merkle_branch(leaves: Sequence[bytes]) -> bytes:
        # BOLT #12:
        # If there are not exactly a power of 2 leaves, then the tree depth will
        # be uneven, with the deepest tree on the lowest-order leaves.
        if len(leaves) == 0:
            # This shouldn't happen, but use a distinctive all-zero value if it does.
            return bytes(32)
        if len(leaves) == 1:
            return leaves[0]

        # Split into order-of-two and remainder.
        branchsize = 1
        while branchsize * 2 < len(leaves):
            branchsize *= 2
        return Bolt12.merkle_pair(Bolt12.merkle_branch(leaves[:branchsize]),
                                  Bolt12.merkle_branch(leaves[branchsize:]))

    def merkle(self) -> bytes:
        ordered = tlv_ordered(self.tlv_table, self.values, self.unknowns, exclude_sigs=True)
        # BOLT #12:
        # The Merkle tree's leaves are, in TLV-ascending order for each tlv:
        # 1. The H(`LnLeaf`,tlv).
        lnleaf = bolt12_h(b'LnLeaf')
        # BOLT #12:
        # 2. The H(`LnAll`|all-tlvs,tlv)
        lnall = self.lnall_ctx(ordered)

        leaves = []
        for num, bytestr in ordered:
            tlv = tlv_enc(num, bytestr)
            leafnode = lnleaf.copy()
            leafnode.update(tlv)
            lnallnode = lnall.copy()
            lnallnode.update(tlv)
            leaves.append(self.merkle_pair(leafnode.digest(), lnallnode.digest()))

        return self.merkle_branch(leaves)

    @staticmethod
    def _check_sig(msgname: str, fieldname: str, merkle_root: bytes, pubkey32: bytes, bip340sig: bytes) -> bool:
        # BOLT #12:
        # Each form is signed using one or more TLV signature elements; TLV
        # types 240 through 1000 are considered signature elements.  For these
        # the tag is `lightning` | `messagename` | `fieldname`, and `msg` is
        # the Merkle-root; `lightning` is the literal 9-byte ASCII string,
        # `messagename` is the name of the TLV stream being signed
        # (i.e. `offer`, `invoice_request` or `invoice`) and the `fieldname`
        # is the TLV field containing the signature (e.g. `signature` or
        # `payer_signature`).

        # BOLT #12:
        # Thus we define ... SIG(`tag`,`msg`,`key`) as the signature of H(`tag`,`msg`) using `key`.
        hash = bolt12_h(bytes('lightning' + msgname + fieldname, encoding='utf8'), merkle_root).digest()
        return verify_schnorr(pubkey32, bip340sig, hash)

    @staticmethod
    def _sign(msgname: str, fieldname: str, merkle_root: bytes, privkey: bytes, bip340sig: bytes) -> bool:
        hash = bolt12_h(bytes('lightning' + msgname + fieldname, encoding='utf8'), merkle_root).digest()
        return sign_schnorr(privkey, hash)

    # BOLT #9:
    # | 8/9   | `var_onion_optin` ... | IN9 |...
    # | 14/15 | `payment_secret` ... | IN9 |...
    # | 16/17 | `basic_mpp` ... | IN9 |
    known_features = {8: 'var_onion_optin',
                      14: 'payment_secret',
                      16: 'basic_mpp'}

    @staticmethod
    def check_features(featureset: bytes) -> Optional[str]:
        """Returns None if OK, otherwise the complaint"""
        # BOLT #12
        # - if `features` contains unknown _even_ bits that are non-zero:
        #     - MUST NOT respond to the offer.
        #     - SHOULD indicate the unknown bit to the user.
        for i in range(0, len(featureset) * 8, 2):
            # Big-endian bitfields are the *worst*
            byte = int(featureset[len(featureset) - 1 - i // 8])
            if byte & (1 << (i % 8)) and i not in Bolt12.known_features:
                return "Unsupported required feature {}".format(i)
        return None


class Offer(Bolt12):
    """Class for an offer"""
    def __init__(self, hrp: str, bytestr: bytes):
        super().__init__(hrp, tlv_offer, bytestr)
        self.offer_id = self.merkle()

    def check(self) -> Tuple[bool, str]:
        """Check it's OK: returns (True, '') or (False, reason)"""
        # BOLT #12:
        #   - if `features` contains unknown _odd_ bits that are non-zero:
        #     - MUST ignore the bit.
        #   - if `features` contains unknown _even_ bits that are non-zero:
        #     - MUST NOT respond to the offer.
        #     - SHOULD indicate the unknown bit to the user.
        whybad = self.check_features(self.values.get('features', bytes()))
        if whybad:
            return False, whybad
        # BOLT #12:
        #   - if `node_id` or `description` is not set:
        #     - MUST NOT respond to the offer.
        if 'node_id' not in self.values:
            return False, "Missing node_id"
        if 'description' not in self.values:
            return False, "Missing description"
        # BOLT #12:
        #  - if `signature` is present, but is not a valid signature
        #    using `node_id` as described in [Signature
        #    Calculation](#signature-calculation): - MUST NOT respond
        #    to the offer.
        if 'signature' in self.values:
            if not self._check_sig("offer", "signature",
                                   self.offer_id,
                                   self.values['node_id'],
                                   self.values['signature']):
                return False, "Bad signature"

        # BOLT #12:
        #  - SHOULD not respond to an offer if the current time is after
        #    `absolute_expiry`.
        if 'absolute_expiry' in self.values and time.time() > self.values['absolute_expiry']:
            return False, "Expired {} seconds ago".format(int(time.time())
                                                          - self.values['absolute_expiry'])

        # BOLT #12:
        #  - FIXME: more!
        return True, ''

    def get_recurrence(self) -> Optional[Recurrence]:
        """If this offer recurs, returns the details.  Otherwise, None"""
        # BOLT #12:
        #   - MAY include `recurrence` to indicate offer should trigger
        #     time-spaced invoices.
        if 'recurrence' not in self.values:
            return None
        return Recurrence(self.values['recurrence'],
                          self.values.get('recurrence_paywindow'),
                          self.values.get('recurrence_limit'),
                          self.values.get('recurrence_base'))


class InvoiceRequest(Bolt12):
    """Class for an invoice_request"""
    def __init__(self, hrp: str, bytestr: bytes):
        super().__init__(hrp, tlv_invoice_request, bytestr)


class Invoice(Bolt12):
    """Class for an invoice"""
    def __init__(self, hrp: str, bytestr: bytes):
        super().__init__(hrp, tlv_invoice, bytestr)


class Decoder(object):
    """Generates a bolt12, or a failure string"""
    def __init__(self):
        self.so_far = ''

    def complete(self) -> bool:
        """Is this string done?"""
        return len(self.so_far) > 0 and not re.search(r'\+\s*$', self.so_far)

    def add(self, bech32str: str):
        """Add this to the string, return complete()"""
        self.so_far += bech32str
        return self.complete()

    def raw_decode(self) -> Tuple[str, bytes]:
        """One string is complete(), try converting to hrp, bytes"""
        assert self.complete()
        if self.so_far.startswith('+'):
            raise ValueError("Missing a part?")

        return simple_bech32_decode(re.sub(r'([A-Z0-9a-z])\+\s*([A-Z0-9a-z])', r'\1\2',
                                           self.so_far))

    def result(self) -> Tuple[Optional[Bolt12], str]:
        """One string is complete(), try decoding"""
        try:
            hrp, bytestr = self.raw_decode()
        except ValueError as e:
            return None, ' '.join(e.args)

        try:
            ret = Bolt12.create(hrp, bytestr)
        except ValueError as e:
            return None, ' '.join(e.args)

        return ret, ''



# From https://github.com/bitcoin/bitcoin/blob/master/test/functional/test_framework/key.py commit f64adc1eedff9d342b49d7e6428b2da21130c23c

# Copyright (c) 2019-2020 Pieter Wuille
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Test-only secp256k1 elliptic curve implementation

WARNING: This code is slow, uses bad randomness, does not properly protect
keys, and is trivially vulnerable to side channel attacks. Do not use for
anything but tests."""
import csv
import hashlib
import os
import random
import unittest

def modinv(a, n):
    """Compute the modular inverse of a modulo n using the extended Euclidean
    Algorithm. See https://en.wikipedia.org/wiki/Extended_Euclidean_algorithm#Modular_integers.
    """
    # TODO: Change to pow(a, -1, n) available in Python 3.8
    t1, t2 = 0, 1
    r1, r2 = n, a
    while r2 != 0:
        q = r1 // r2
        t1, t2 = t2, t1 - q * t2
        r1, r2 = r2, r1 - q * r2
    if r1 > 1:
        return None
    if t1 < 0:
        t1 += n
    return t1

def TaggedHash(tag, data):
    ss = hashlib.sha256(tag.encode('utf-8')).digest()
    ss += ss
    ss += data
    return hashlib.sha256(ss).digest()

def jacobi_symbol(n, k):
    """Compute the Jacobi symbol of n modulo k

    See https://en.wikipedia.org/wiki/Jacobi_symbol

    For our application k is always prime, so this is the same as the Legendre symbol."""
    assert k > 0 and k & 1, "jacobi symbol is only defined for positive odd k"
    n %= k
    t = 0
    while n != 0:
        while n & 1 == 0:
            n >>= 1
            r = k & 7
            t ^= (r == 3 or r == 5)
        n, k = k, n
        t ^= (n & k & 3 == 3)
        n = n % k
    if k == 1:
        return -1 if t else 1
    return 0

def modsqrt(a, p):
    """Compute the square root of a modulo p when p % 4 = 3.

    The Tonelli-Shanks algorithm can be used. See https://en.wikipedia.org/wiki/Tonelli-Shanks_algorithm

    Limiting this function to only work for p % 4 = 3 means we don't need to
    iterate through the loop. The highest n such that p - 1 = 2^n Q with Q odd
    is n = 1. Therefore Q = (p-1)/2 and sqrt = a^((Q+1)/2) = a^((p+1)/4)

    secp256k1's is defined over field of size 2**256 - 2**32 - 977, which is 3 mod 4.
    """
    if p % 4 != 3:
        raise NotImplementedError("modsqrt only implemented for p % 4 = 3")
    sqrt = pow(a, (p + 1)//4, p)
    if pow(sqrt, 2, p) == a % p:
        return sqrt
    return None

class EllipticCurve:
    def __init__(self, p, a, b):
        """Initialize elliptic curve y^2 = x^3 + a*x + b over GF(p)."""
        self.p = p
        self.a = a % p
        self.b = b % p

    def affine(self, p1):
        """Convert a Jacobian point tuple p1 to affine form, or None if at infinity.

        An affine point is represented as the Jacobian (x, y, 1)"""
        x1, y1, z1 = p1
        if z1 == 0:
            return None
        inv = modinv(z1, self.p)
        inv_2 = (inv**2) % self.p
        inv_3 = (inv_2 * inv) % self.p
        return ((inv_2 * x1) % self.p, (inv_3 * y1) % self.p, 1)

    def has_even_y(self, p1):
        """Whether the point p1 has an even Y coordinate when expressed in affine coordinates."""
        return not (p1[2] == 0 or self.affine(p1)[1] & 1)

    def negate(self, p1):
        """Negate a Jacobian point tuple p1."""
        x1, y1, z1 = p1
        return (x1, (self.p - y1) % self.p, z1)

    def on_curve(self, p1):
        """Determine whether a Jacobian tuple p is on the curve (and not infinity)"""
        x1, y1, z1 = p1
        z2 = pow(z1, 2, self.p)
        z4 = pow(z2, 2, self.p)
        return z1 != 0 and (pow(x1, 3, self.p) + self.a * x1 * z4 + self.b * z2 * z4 - pow(y1, 2, self.p)) % self.p == 0

    def is_x_coord(self, x):
        """Test whether x is a valid X coordinate on the curve."""
        x_3 = pow(x, 3, self.p)
        return jacobi_symbol(x_3 + self.a * x + self.b, self.p) != -1

    def lift_x(self, x):
        """Given an X coordinate on the curve, return a corresponding affine point for which the Y coordinate is even."""
        x_3 = pow(x, 3, self.p)
        v = x_3 + self.a * x + self.b
        y = modsqrt(v, self.p)
        if y is None:
            return None
        return (x, self.p - y if y & 1 else y, 1)

    def double(self, p1):
        """Double a Jacobian tuple p1

        See https://en.wikibooks.org/wiki/Cryptography/Prime_Curve/Jacobian_Coordinates - Point Doubling"""
        x1, y1, z1 = p1
        if z1 == 0:
            return (0, 1, 0)
        y1_2 = (y1**2) % self.p
        y1_4 = (y1_2**2) % self.p
        x1_2 = (x1**2) % self.p
        s = (4*x1*y1_2) % self.p
        m = 3*x1_2
        if self.a:
            m += self.a * pow(z1, 4, self.p)
        m = m % self.p
        x2 = (m**2 - 2*s) % self.p
        y2 = (m*(s - x2) - 8*y1_4) % self.p
        z2 = (2*y1*z1) % self.p
        return (x2, y2, z2)

    def add_mixed(self, p1, p2):
        """Add a Jacobian tuple p1 and an affine tuple p2

        See https://en.wikibooks.org/wiki/Cryptography/Prime_Curve/Jacobian_Coordinates - Point Addition (with affine point)"""
        x1, y1, z1 = p1
        x2, y2, z2 = p2
        assert(z2 == 1)
        # Adding to the point at infinity is a no-op
        if z1 == 0:
            return p2
        z1_2 = (z1**2) % self.p
        z1_3 = (z1_2 * z1) % self.p
        u2 = (x2 * z1_2) % self.p
        s2 = (y2 * z1_3) % self.p
        if x1 == u2:
            if (y1 != s2):
                # p1 and p2 are inverses. Return the point at infinity.
                return (0, 1, 0)
            # p1 == p2. The formulas below fail when the two points are equal.
            return self.double(p1)
        h = u2 - x1
        r = s2 - y1
        h_2 = (h**2) % self.p
        h_3 = (h_2 * h) % self.p
        u1_h_2 = (x1 * h_2) % self.p
        x3 = (r**2 - h_3 - 2*u1_h_2) % self.p
        y3 = (r*(u1_h_2 - x3) - y1*h_3) % self.p
        z3 = (h*z1) % self.p
        return (x3, y3, z3)

    def add(self, p1, p2):
        """Add two Jacobian tuples p1 and p2

        See https://en.wikibooks.org/wiki/Cryptography/Prime_Curve/Jacobian_Coordinates - Point Addition"""
        x1, y1, z1 = p1
        x2, y2, z2 = p2
        # Adding the point at infinity is a no-op
        if z1 == 0:
            return p2
        if z2 == 0:
            return p1
        # Adding an Affine to a Jacobian is more efficient since we save field multiplications and squarings when z = 1
        if z1 == 1:
            return self.add_mixed(p2, p1)
        if z2 == 1:
            return self.add_mixed(p1, p2)
        z1_2 = (z1**2) % self.p
        z1_3 = (z1_2 * z1) % self.p
        z2_2 = (z2**2) % self.p
        z2_3 = (z2_2 * z2) % self.p
        u1 = (x1 * z2_2) % self.p
        u2 = (x2 * z1_2) % self.p
        s1 = (y1 * z2_3) % self.p
        s2 = (y2 * z1_3) % self.p
        if u1 == u2:
            if (s1 != s2):
                # p1 and p2 are inverses. Return the point at infinity.
                return (0, 1, 0)
            # p1 == p2. The formulas below fail when the two points are equal.
            return self.double(p1)
        h = u2 - u1
        r = s2 - s1
        h_2 = (h**2) % self.p
        h_3 = (h_2 * h) % self.p
        u1_h_2 = (u1 * h_2) % self.p
        x3 = (r**2 - h_3 - 2*u1_h_2) % self.p
        y3 = (r*(u1_h_2 - x3) - s1*h_3) % self.p
        z3 = (h*z1*z2) % self.p
        return (x3, y3, z3)

    def mul(self, ps):
        """Compute a (multi) point multiplication

        ps is a list of (Jacobian tuple, scalar) pairs.
        """
        r = (0, 1, 0)
        for i in range(255, -1, -1):
            r = self.double(r)
            for (p, n) in ps:
                if ((n >> i) & 1):
                    r = self.add(r, p)
        return r

SECP256K1_FIELD_SIZE = 2**256 - 2**32 - 977
SECP256K1 = EllipticCurve(SECP256K1_FIELD_SIZE, 0, 7)
SECP256K1_G = (0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798, 0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8, 1)
SECP256K1_ORDER = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
SECP256K1_ORDER_HALF = SECP256K1_ORDER // 2

class ECPubKey():
    """A secp256k1 public key"""

    def __init__(self):
        """Construct an uninitialized public key"""
        self.valid = False

    def set(self, data):
        """Construct a public key from a serialization in compressed or uncompressed format"""
        if (len(data) == 65 and data[0] == 0x04):
            p = (int.from_bytes(data[1:33], 'big'), int.from_bytes(data[33:65], 'big'), 1)
            self.valid = SECP256K1.on_curve(p)
            if self.valid:
                self.p = p
                self.compressed = False
        elif (len(data) == 33 and (data[0] == 0x02 or data[0] == 0x03)):
            x = int.from_bytes(data[1:33], 'big')
            if SECP256K1.is_x_coord(x):
                p = SECP256K1.lift_x(x)
                # Make the Y coordinate odd if required (lift_x always produces
                # a point with an even Y coordinate).
                if data[0] & 1:
                    p = SECP256K1.negate(p)
                self.p = p
                self.valid = True
                self.compressed = True
            else:
                self.valid = False
        else:
            self.valid = False

    @property
    def is_compressed(self):
        return self.compressed

    @property
    def is_valid(self):
        return self.valid

    def get_bytes(self):
        assert(self.valid)
        p = SECP256K1.affine(self.p)
        if p is None:
            return None
        if self.compressed:
            return bytes([0x02 + (p[1] & 1)]) + p[0].to_bytes(32, 'big')
        else:
            return bytes([0x04]) + p[0].to_bytes(32, 'big') + p[1].to_bytes(32, 'big')

    def verify_ecdsa(self, sig, msg, low_s=True):
        """Verify a strictly DER-encoded ECDSA signature against this pubkey.

        See https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm for the
        ECDSA verifier algorithm"""
        assert(self.valid)

        # Extract r and s from the DER formatted signature. Return false for
        # any DER encoding errors.
        if (sig[1] + 2 != len(sig)):
            return False
        if (len(sig) < 4):
            return False
        if (sig[0] != 0x30):
            return False
        if (sig[2] != 0x02):
            return False
        rlen = sig[3]
        if (len(sig) < 6 + rlen):
            return False
        if rlen < 1 or rlen > 33:
            return False
        if sig[4] >= 0x80:
            return False
        if (rlen > 1 and (sig[4] == 0) and not (sig[5] & 0x80)):
            return False
        r = int.from_bytes(sig[4:4+rlen], 'big')
        if (sig[4+rlen] != 0x02):
            return False
        slen = sig[5+rlen]
        if slen < 1 or slen > 33:
            return False
        if (len(sig) != 6 + rlen + slen):
            return False
        if sig[6+rlen] >= 0x80:
            return False
        if (slen > 1 and (sig[6+rlen] == 0) and not (sig[7+rlen] & 0x80)):
            return False
        s = int.from_bytes(sig[6+rlen:6+rlen+slen], 'big')

        # Verify that r and s are within the group order
        if r < 1 or s < 1 or r >= SECP256K1_ORDER or s >= SECP256K1_ORDER:
            return False
        if low_s and s >= SECP256K1_ORDER_HALF:
            return False
        z = int.from_bytes(msg, 'big')

        # Run verifier algorithm on r, s
        w = modinv(s, SECP256K1_ORDER)
        u1 = z*w % SECP256K1_ORDER
        u2 = r*w % SECP256K1_ORDER
        R = SECP256K1.affine(SECP256K1.mul([(SECP256K1_G, u1), (self.p, u2)]))
        if R is None or (R[0] % SECP256K1_ORDER) != r:
            return False
        return True

def generate_privkey():
    """Generate a valid random 32-byte private key."""
    return random.randrange(1, SECP256K1_ORDER).to_bytes(32, 'big')

class ECKey():
    """A secp256k1 private key"""

    def __init__(self):
        self.valid = False

    def set(self, secret, compressed):
        """Construct a private key object with given 32-byte secret and compressed flag."""
        assert(len(secret) == 32)
        secret = int.from_bytes(secret, 'big')
        self.valid = (secret > 0 and secret < SECP256K1_ORDER)
        if self.valid:
            self.secret = secret
            self.compressed = compressed

    def generate(self, compressed=True):
        """Generate a random private key (compressed or uncompressed)."""
        self.set(generate_privkey(), compressed)

    def get_bytes(self):
        """Retrieve the 32-byte representation of this key."""
        assert(self.valid)
        return self.secret.to_bytes(32, 'big')

    @property
    def is_valid(self):
        return self.valid

    @property
    def is_compressed(self):
        return self.compressed

    def get_pubkey(self):
        """Compute an ECPubKey object for this secret key."""
        assert(self.valid)
        ret = ECPubKey()
        p = SECP256K1.mul([(SECP256K1_G, self.secret)])
        ret.p = p
        ret.valid = True
        ret.compressed = self.compressed
        return ret

    def sign_ecdsa(self, msg, low_s=True):
        """Construct a DER-encoded ECDSA signature with this key.

        See https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm for the
        ECDSA signer algorithm."""
        assert(self.valid)
        z = int.from_bytes(msg, 'big')
        # Note: no RFC6979, but a simple random nonce (some tests rely on distinct transactions for the same operation)
        k = random.randrange(1, SECP256K1_ORDER)
        R = SECP256K1.affine(SECP256K1.mul([(SECP256K1_G, k)]))
        r = R[0] % SECP256K1_ORDER
        s = (modinv(k, SECP256K1_ORDER) * (z + self.secret * r)) % SECP256K1_ORDER
        if low_s and s > SECP256K1_ORDER_HALF:
            s = SECP256K1_ORDER - s
        # Represent in DER format. The byte representations of r and s have
        # length rounded up (255 bits becomes 32 bytes and 256 bits becomes 33
        # bytes).
        rb = r.to_bytes((r.bit_length() + 8) // 8, 'big')
        sb = s.to_bytes((s.bit_length() + 8) // 8, 'big')
        return b'\x30' + bytes([4 + len(rb) + len(sb), 2, len(rb)]) + rb + bytes([2, len(sb)]) + sb

def compute_xonly_pubkey(key):
    """Compute an x-only (32 byte) public key from a (32 byte) private key.

    This also returns whether the resulting public key was negated.
    """

    assert len(key) == 32
    x = int.from_bytes(key, 'big')
    if x == 0 or x >= SECP256K1_ORDER:
        return (None, None)
    P = SECP256K1.affine(SECP256K1.mul([(SECP256K1_G, x)]))
    return (P[0].to_bytes(32, 'big'), not SECP256K1.has_even_y(P))

def tweak_add_privkey(key, tweak):
    """Tweak a private key (after negating it if needed)."""

    assert len(key) == 32
    assert len(tweak) == 32

    x = int.from_bytes(key, 'big')
    if x == 0 or x >= SECP256K1_ORDER:
        return None
    if not SECP256K1.has_even_y(SECP256K1.mul([(SECP256K1_G, x)])):
       x = SECP256K1_ORDER - x
    t = int.from_bytes(tweak, 'big')
    if t >= SECP256K1_ORDER:
        return None
    x = (x + t) % SECP256K1_ORDER
    if x == 0:
        return None
    return x.to_bytes(32, 'big')

def tweak_add_pubkey(key, tweak):
    """Tweak a public key and return whether the result had to be negated."""

    assert len(key) == 32
    assert len(tweak) == 32

    x_coord = int.from_bytes(key, 'big')
    if x_coord >= SECP256K1_FIELD_SIZE:
        return None
    P = SECP256K1.lift_x(x_coord)
    if P is None:
        return None
    t = int.from_bytes(tweak, 'big')
    if t >= SECP256K1_ORDER:
        return None
    Q = SECP256K1.affine(SECP256K1.mul([(SECP256K1_G, t), (P, 1)]))
    if Q is None:
        return None
    return (Q[0].to_bytes(32, 'big'), not SECP256K1.has_even_y(Q))

def verify_schnorr(key, sig, msg):
    """Verify a Schnorr signature (see BIP 340).

    - key is a 32-byte xonly pubkey (computed using compute_xonly_pubkey).
    - sig is a 64-byte Schnorr signature
    - msg is a 32-byte message
    """
    assert len(key) == 32
    assert len(msg) == 32
    assert len(sig) == 64

    x_coord = int.from_bytes(key, 'big')
    if x_coord == 0 or x_coord >= SECP256K1_FIELD_SIZE:
        return False
    P = SECP256K1.lift_x(x_coord)
    if P is None:
        return False
    r = int.from_bytes(sig[0:32], 'big')
    if r >= SECP256K1_FIELD_SIZE:
        return False
    s = int.from_bytes(sig[32:64], 'big')
    if s >= SECP256K1_ORDER:
        return False
    e = int.from_bytes(TaggedHash("BIP0340/challenge", sig[0:32] + key + msg), 'big') % SECP256K1_ORDER
    R = SECP256K1.mul([(SECP256K1_G, s), (P, SECP256K1_ORDER - e)])
    if not SECP256K1.has_even_y(R):
        return False
    if ((r * R[2] * R[2]) % SECP256K1_FIELD_SIZE) != R[0]:
        return False
    return True

def sign_schnorr(key, msg, aux=None, flip_p=False, flip_r=False):
    """Create a Schnorr signature (see BIP 340)."""

    if aux is None:
        aux = bytes(32)

    assert len(key) == 32
    assert len(msg) == 32
    assert len(aux) == 32

    sec = int.from_bytes(key, 'big')
    if sec == 0 or sec >= SECP256K1_ORDER:
        return None
    P = SECP256K1.affine(SECP256K1.mul([(SECP256K1_G, sec)]))
    if SECP256K1.has_even_y(P) == flip_p:
        sec = SECP256K1_ORDER - sec
    t = (sec ^ int.from_bytes(TaggedHash("BIP0340/aux", aux), 'big')).to_bytes(32, 'big')
    kp = int.from_bytes(TaggedHash("BIP0340/nonce", t + P[0].to_bytes(32, 'big') + msg), 'big') % SECP256K1_ORDER
    assert kp != 0
    R = SECP256K1.affine(SECP256K1.mul([(SECP256K1_G, kp)]))
    k = kp if SECP256K1.has_even_y(R) != flip_r else SECP256K1_ORDER - kp
    e = int.from_bytes(TaggedHash("BIP0340/challenge", R[0].to_bytes(32, 'big') + P[0].to_bytes(32, 'big') + msg), 'big') % SECP256K1_ORDER
    return R[0].to_bytes(32, 'big') + ((k + e * sec) % SECP256K1_ORDER).to_bytes(32, 'big')

class TestFrameworkKey(unittest.TestCase):
    def test_schnorr(self):
        """Test the Python Schnorr implementation."""
        byte_arrays = [generate_privkey() for _ in range(3)] + [v.to_bytes(32, 'big') for v in [0, SECP256K1_ORDER - 1, SECP256K1_ORDER, 2**256 - 1]]
        keys = {}
        for privkey in byte_arrays:  # build array of key/pubkey pairs
            pubkey, _ = compute_xonly_pubkey(privkey)
            if pubkey is not None:
                keys[privkey] = pubkey
        for msg in byte_arrays:  # test every combination of message, signing key, verification key
            for sign_privkey, _ in keys.items():
                sig = sign_schnorr(sign_privkey, msg)
                for verify_privkey, verify_pubkey in keys.items():
                    if verify_privkey == sign_privkey:
                        self.assertTrue(verify_schnorr(verify_pubkey, sig, msg))
                        sig = list(sig)
                        sig[random.randrange(64)] ^= (1 << (random.randrange(8)))  # damaging signature should break things
                        sig = bytes(sig)
                    self.assertFalse(verify_schnorr(verify_pubkey, sig, msg))

    def test_schnorr_testvectors(self):
        """Implement the BIP340 test vectors (read from bip340_test_vectors.csv)."""
        num_tests = 0
        vectors_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'bip340_test_vectors.csv')
        with open(vectors_file, newline='', encoding='utf8') as csvfile:
            reader = csv.reader(csvfile)
            next(reader)
            for row in reader:
                (i_str, seckey_hex, pubkey_hex, aux_rand_hex, msg_hex, sig_hex, result_str, comment) = row
                i = int(i_str)
                pubkey = bytes.fromhex(pubkey_hex)
                msg = bytes.fromhex(msg_hex)
                sig = bytes.fromhex(sig_hex)
                result = result_str == 'TRUE'
                if seckey_hex != '':
                    seckey = bytes.fromhex(seckey_hex)
                    pubkey_actual = compute_xonly_pubkey(seckey)[0]
                    self.assertEqual(pubkey.hex(), pubkey_actual.hex(), "BIP340 test vector %i (%s): pubkey mismatch" % (i, comment))
                    aux_rand = bytes.fromhex(aux_rand_hex)
                    try:
                        sig_actual = sign_schnorr(seckey, msg, aux_rand)
                        self.assertEqual(sig.hex(), sig_actual.hex(), "BIP340 test vector %i (%s): sig mismatch" % (i, comment))
                    except RuntimeError as e:
                        self.fail("BIP340 test vector %i (%s): signing raised exception %s" % (i, comment, e))
                result_actual = verify_schnorr(pubkey, sig, msg)
                if result:
                    self.assertEqual(result, result_actual, "BIP340 test vector %i (%s): verification failed" % (i, comment))
                else:
                    self.assertEqual(result, result_actual, "BIP340 test vector %i (%s): verification succeeded unexpectedly" % (i, comment))
                num_tests += 1
        self.assertTrue(num_tests >= 15) # expect at least 15 test vectors



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bolt12']

package_data = \
{'': ['*']}

install_requires = \
['bech32>=1.2.0,<2.0.0',
 'pyln-proto>=0.10.1,<0.11.0',
 'python-dateutil>=2.8.2,<3.0.0']

setup_kwargs = {
    'name': 'bolt12',
    'version': '0.1.2',
    'description': 'Lightning Network BOLT12 Routines',
    'long_description': '# Lightning Network BOLT12 Routines\n\nThis is a simple Python 3 library to read, check, generate and\nmanipulate [BOLT 12](https://bolt12.org) offers, invoice_requests and\ninvoices.\n\nSee [examples/](examples/) for usage.\n',
    'author': 'Rusty Russell',
    'author_email': 'rusty@rustcorp.com.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rustyrussell/bolt12/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
